import numpy as np
import pandas as pd

from app import comp
from app.comp.moy_sem import comp_ranks_series
from app.models import UniteEns
from app.pe import pe_affichage
from app.scodoc.codes_cursus import UE_SPORT


class Moyenne:
    CRITERES = [
        "note",
        "classement",
        "rang",
        "min",
        "max",
        "moy",
        "nb_etuds",
        "nb_inscrits",
    ]

    def __init__(self, notes: pd.Series):
        """Classe centralisant la synthèse des moyennes/classements d'une série
        de notes :

        * des "notes": la Serie pandas des notes (float),
        * des "classements": la Serie pandas des classements (float),
        * des "min": la note minimum,
        * des "max": la note maximum,
        * des "moy": la moyenne,
        * des "nb_inscrits": le nombre d'étudiants ayant une note,
        """
        self.notes = notes
        """Les notes"""
        self.etudids = list(notes.index)  # calcul à venir
        """Les id des étudiants"""
        self.inscrits_ids = notes[notes.notnull()].index.to_list()
        """Les id des étudiants dont la note est non nulle"""
        self.df: pd.DataFrame = self.comp_moy_et_stat(self.notes)
        """Le dataframe retraçant les moyennes/classements/statistiques"""
        self.synthese = self.to_dict()
        """La synthèse (dictionnaire) des notes/classements/statistiques"""

    def comp_moy_et_stat(self, notes: pd.Series) -> dict:
        """Calcule et structure les données nécessaires au PE pour une série
        de notes (pouvant être une moyenne d'un tag à une UE ou une moyenne générale
        d'un tag) dans un dictionnaire spécifique.

        Partant des notes, sont calculés les classements (en ne tenant compte
        que des notes non nulles).

        Args:
            notes: Une série de notes (avec des éventuels NaN)

        Returns:
            Un dictionnaire stockant les notes, les classements, le min,
            le max, la moyenne, le nb de notes (donc d'inscrits)
        """
        df = pd.DataFrame(
            np.nan,
            index=self.etudids,
            columns=Moyenne.CRITERES,
        )

        # Supprime d'éventuelles chaines de caractères dans les notes
        notes = pd.to_numeric(notes, errors="coerce")
        df["note"] = notes

        # Les nb d'étudiants & nb d'inscrits
        df["nb_etuds"] = len(self.etudids)
        # Les étudiants dont la note n'est pas nulle
        inscrits_ids = notes[notes.notnull()].index.to_list()
        df.loc[inscrits_ids, "nb_inscrits"] = len(inscrits_ids)

        # Le classement des inscrits
        notes_non_nulles = notes[inscrits_ids]
        (class_str, class_int) = comp_ranks_series(notes_non_nulles)
        df.loc[inscrits_ids, "classement"] = class_int

        # Le rang (classement/nb_inscrit)
        df["rang"] = df["rang"].astype(str)
        df.loc[inscrits_ids, "rang"] = (
            df.loc[inscrits_ids, "classement"].astype(int).astype(str)
            + "/"
            + df.loc[inscrits_ids, "nb_inscrits"].astype(int).astype(str)
        )

        # Les stat (des inscrits)
        df.loc[inscrits_ids, "min"] = notes.min()
        df.loc[inscrits_ids, "max"] = notes.max()
        df.loc[inscrits_ids, "moy"] = notes.mean()

        return df

    def to_dict(self) -> dict:
        """Renvoie un dictionnaire de synthèse des moyennes/classements/statistiques générale (but)"""
        synthese = {
            "notes": self.df["note"],
            "classements": self.df["classement"],
            "min": self.df["min"].mean(),
            "max": self.df["max"].mean(),
            "moy": self.df["moy"].mean(),
            "nb_inscrits": self.df["nb_inscrits"].mean(),
        }
        return synthese

    def get_notes(self):
        """Série des notes, arrondies à 2 chiffres après la virgule"""
        return self.df_gen["note"].round(2)

    def get_rangs_inscrits(self) -> pd.Series:
        """Série des rangs classement/nbre_inscrit"""
        return self.df_gen["rang"]

    def get_min(self) -> pd.Series:
        """Série des min"""
        return self.df_gen["min"].round(2)

    def get_max(self) -> pd.Series:
        """Série des max"""
        return self.df_gen["max"].round(2)

    def get_moy(self) -> pd.Series:
        """Série des moy"""
        return self.df_gen["moy"].round(2)

    def get_note_for_df(self, etudid: int):
        """Note d'un étudiant donné par son etudid"""
        return round(self.df_gen["note"].loc[etudid], 2)

    def get_min_for_df(self) -> float:
        """Min renseigné pour affichage dans un df"""
        return round(self.synthese["min"], 2)

    def get_max_for_df(self) -> float:
        """Max renseigné pour affichage dans un df"""
        return round(self.synthese["max"], 2)

    def get_moy_for_df(self) -> float:
        """Moyenne renseignée pour affichage dans un df"""
        return round(self.synthese["moy"], 2)

    def get_class_for_df(self, etudid: int) -> str:
        """Classement ramené au nombre d'inscrits,
        pour un étudiant donné par son etudid"""
        classement = self.df_gen["rang"].loc[etudid]
        if not pd.isna(classement):
            return classement
        else:
            return pe_affichage.SANS_NOTE

    def is_significatif(self) -> bool:
        """Indique si la moyenne est significative (c'est-à-dire à des notes)"""
        return self.synthese["nb_inscrits"] > 0


class MoyennesTag:
    def __init__(
        self,
        tag: str,
        ues: dict[int, UniteEns],
        notes_ues: pd.DataFrame,
        ues_inscr_parcours_df: pd.DataFrame
        # notes_gen: pd.Series,
    ):
        """Classe centralisant la synthèse des moyennes/classements d'une série
        d'étudiants à un tag donné, en différenciant les notes
        obtenues aux UE et au général (toutes UEs confondues)

        Args:
            tag: Un tag
            ues: La liste des UEs ayant servie au calcul de la moyenne
            notes_ues: Les moyennes (etudid x acronymes_ues) aux différentes UEs et pour le tag
            ues_inscr_parcours_df: Les inscriptions des etudid au UE
            # notes_gen: Une série de notes (moyenne) sous forme d'un pd.Series() (toutes UEs confondues)
        """
        self.tag = tag
        """Le tag associé aux moyennes"""

        # Les UE
        self.ues: dict[int, UniteEns] = ues
        """Les UEs sur lesquelles sont calculées les moyennes"""

        colonnes = list(notes_ues.columns)
        acronymes: list[str] = [self.ues[ue_id].acronyme for ue_id in self.ues]
        assert len(set(acronymes)) == len(
            colonnes
        ), "Deux UEs ne peuvent pas avoir le même acronyme"

        # Les inscriptions des etudids aux UEs
        self.ues_inscr_parcours_df: pd.DataFrame = ues_inscr_parcours_df
        """Les inscriptions des etudids au UE en fonction de leur parcours"""

        # Les coefficients à appliquer aux UEs pour la moyenne générale = ECTS
        self.ects = self.ues_inscr_parcours_df.fillna(0.0) * [
            ue.ects
            for ue in self.ues.values()  # if ue.type != UE_SPORT <= déjà supprimé
        ]
        # Les profils d'ects (pour debug)
        profils_ects = []
        for val in list(self.ects.values):
            if tuple(val) not in profils_ects:
                profils_ects.append(tuple(val))

        # Les moyennes par UE
        self.notes_ues: pd.DataFrame = notes_ues
        """Les notes aux UEs (dataframe)"""
        self.notes_ues.columns = acronymes  # remplace les ue.id par leur acronyme
        self.moys_ues: dict[int, pd.DataFrame] = {}
        """Les dataframes retraçant les moyennes/classements/statistiques des étudiants aux UEs"""
        for ue in self.ues.values():  # if ue.type != UE_SPORT:
            notes = notes_ues[ue.acronyme]
            self.moys_ues[ue.acronyme] = Moyenne(notes)

        # Les moyennes générales
        notes_gen = self.compute_moy_gen(self.notes_ues, self.ects)
        self.notes_gen = notes_gen
        """Les notes générales (moyenne toutes UEs confonudes)"""
        self.moy_gen = Moyenne(notes_gen)
        """Le dataframe retraçant les moyennes/classements/statistiques général"""

        pe_affichage.pe_print(f"> MoyTag pour {tag} avec")
        pe_affichage.pe_print(f"  - ues={acronymes}")
        pe_affichage.pe_print(f"  - ects={profils_ects}")


    def __eq__(self, other):
        """Egalité de deux MoyenneTag lorsque leur tag sont identiques"""
        return self.tag == other.tag

    def compute_moy_gen(
        self, moy_ues: pd.DataFrame, coeff_ues: pd.DataFrame
    ) -> pd.Series:
        """Calcule la moyenne générale (toutes UE confondus)
        pour le tag considéré, en pondérant les notes obtenues au UE
        par les crédits ECTS.

        Args:
            moy_ues: Les moyennes etudids x acronymes_ues
            coeff_ues: Les coeff etudids x ueids
        """

        # Calcule la moyenne générale dans le semestre (pondérée par le ECTS)
        try:
            moy_gen_tag = comp.moy_sem.compute_sem_moys_apc_using_ects(
                moy_ues,
                coeff_ues,
                # formation_id=self.formsemestre.formation_id,
                skip_empty_ues=True,
            )
        except TypeError as e:
            raise TypeError("Pb dans le calcul de la moyenne toutes UEs confondues")

        return moy_gen_tag
